@@ -23,13 +23,16 @@
 #include "../common/util.h"
 #include "../common/strlist.h"
 
+
 /* We keep all global options in the structure OPT.  */
+EXTERN_UNLESS_MAIN_MODULE
 struct
 {
   int verbose;
   unsigned int debug_level;
   int quiet;
   int dry_run;
+  int utf8strings;
   const char *gpg_program;
   strlist_t gpg_arguments;
   const char *outfile;
@@ -92,6 +95,8 @@ typedef enum
     TF_DIRECTORY,
     TF_FIFO,
     TF_RESERVED,
+    TF_GEXTHDR,    /* Global extended header.  */
+    TF_EXTHDR,     /* Extended header.  */
     TF_UNKNOWN,    /* Needs to be treated as regular file.  */
     TF_NOTSUP      /* Not supported (used with --create).  */
   } typeflag_t;
@@ -119,7 +124,7 @@ struct tar_header_s
 
   unsigned long long nrecords; /* Number of data records.  */
 
-  char name[1];             /* Filename (dynamically extended).  */
+  char name[1];             /* Filename (UTF-8, dynamically extended).  */
 };
 
 
@@ -128,7 +133,8 @@ gpg_error_t read_record (estream_t stream, void *record);
 gpg_error_t write_record (estream_t stream, const void *record);
 
 /*-- gpgtar-create.c --*/
-gpg_error_t gpgtar_create (char **inpattern, int encrypt, int sign);
+gpg_error_t gpgtar_create (char **inpattern, const char *files_from,
+                           int null_names, int encrypt, int sign);
 
 /*-- gpgtar-extract.c --*/
 gpg_error_t gpgtar_extract (const char *filename, int decrypt);
@@ -136,8 +142,9 @@ gpg_error_t gpgtar_extract (const char *filename, int decrypt);
 /*-- gpgtar-list.c --*/
 gpg_error_t gpgtar_list (const char *filename, int decrypt);
 gpg_error_t gpgtar_read_header (estream_t stream, tarinfo_t info,
-                                tar_header_t *r_header);
-void gpgtar_print_header (tar_header_t header, estream_t out);
+                                tar_header_t *r_header, strlist_t *r_extheader);
+void gpgtar_print_header (tar_header_t header, strlist_t extheader,
+                          estream_t out);
 
 
 #endif /*GPGTAR_H*/
